const { contextBridge, ipcRenderer } = require('electron');
const { networkInterfaces, tmpdir } = require('os');
const { sep, join } = require('path');
const fs = require('fs');
const { performance } = require('perf_hooks');
const { get } = require('http');

function promisify(fun) {
  return function (...args) {
    return new Promise((resolve, reject) => {
      fun(...args, (err, extra) => {
        if (err) reject(err);
        else {
          resolve(extra);
        }
      });
    });
  };
}

const rmdir = promisify(fs.rmdir);
const mkdtemp = promisify(fs.mkdtemp);
const readFile = promisify(fs.readFile);
const writeFile = promisify(fs.writeFile);
const stat = promisify(fs.stat);

function getInterfaces() {
  return Array.from(Object.values(networkInterfaces()))
    .flat()
    .filter((interface) => {
      return interface.family === 'IPv4' && !interface.internal;
    });
}

async function getHTTPUrl() {
  const interfaces = getInterfaces();

  if (interfaces.length) {
    return 'http://' + interfaces[0].address + ':' +
      await ipcRenderer.invoke('GetHTTPPort');
  }

  return undefined;
}

function mkTempDir() {
  const tmpDir = tmpdir();

  return mkdtemp(`${tmpDir}${sep}`);
}

function delay(n) {
  return new Promise(function (resolve) {
    setTimeout(resolve, n);
  });
}

async function createTmpFileLocation(ext) {
  let dir = await mkTempDir();

  if (ext && (ext.search('/') !== -1 || ext.search('\\.') !== -1))
    throw new TypeError('Invalid extension.');

  const fname = join(dir, ext ? `file.${ext}` : 'file');

  return {
    path: fname,
    dispose: function () {
      if (dir === null) return;
      const tmp = dir;
      dir = null;
      return rmdir(tmp, {
        recursive: true,
        retryDelay: 500,
        maxRetries: 15,
      });
    },
    read: async function () {
      if (dir === null) throw new Error('dispose() has already been called.');
      const buffer = await readFile(fname, { encoding: null });
      return buffer.buffer;
    },
    write: function (buffer) {
      return writeFile(fname, Buffer.from(buffer), {
        encoding: null,
      });
    },
    wait: async function (timeout) {
      let sameSize = 0;
      let lastSize = -1;
      let startTime = performance.now();
      if (!timeout) timeout = 5000;

      for (;;) {
        try {
          await delay(200);
          const size = (await stat(fname)).size;

          if (lastSize === size) {
            sameSize++;
          } else {
            sameSize = 0;
            lastSize = size;
          }
        } catch (err) {
          sameSize = 0;
          lastSize = -1;
          await delay(200);
        }

        if (lastSize > 0 && sameSize >= 3) {
          // The file has not changed in size for 600ms.
          return;
        }

        if (performance.now() - startTime > timeout)
          throw new Error('Timeout.');
      }
    },
  };
}

async function storeToTmpFile(data) {
  const result = await createTmpFileLocation();
  const fname = result.path;

  try {
    await result.write(data);
  } catch (err) {
    throw err;
  } finally {
    await result.dispose();
  }

  return result;
}

function fetchToken(url) {
  return new Promise((resolve, reject) => {
    if (!url.endsWith('/_control_token')) {
      reject(new Error('Bad token URL.'));
    } else {
      get(url, { timeout: 1000 }, (res) => {
        if (res.statusCode !== 200) {
          reject(new Error('Loading token failed.'));
        } else {
          res.setEncoding('utf8');
          let rawData = '';
          res.on('data', (chunk) => {
            rawData += chunk;
          });
          res.on('end', () => {
            resolve(rawData);
          });
        }
      }).on('error', reject);
    }
  });
}

function GoOnline(localIP, remoteIP, direction) {
  return ipcRenderer.invoke('GoOnline', localIP, remoteIP, direction);
}

function GoOffline() {
  return ipcRenderer.invoke('GoOffline');
}

contextBridge.exposeInMainWorld('yhostAPI', {
  getInterfaces,
  getHTTPUrl,
  storeToTmpFile,
  createTmpFileLocation,
  fetchToken,
  GoOnline,
  GoOffline,
});
